import numpy as np
from scipy import linalg
import matplotlib.pyplot as plt
from mne.time_frequency import psd_array_welch
from mne.viz import plot_topomap
import mne
import seaborn as sns
from .misc import _get_ixs_goods
from mne.filter import next_fast_len
from scipy.signal import hilbert
from scipy.stats import binomtest
from pycircstat.descriptive import mean as circmean
from .base import RawCLAM, EpochsCLAM


def _get_lcmv_weights(COV, forward):
    COVinv = linalg.pinv(COV)
    return ((COVinv @ forward[:, None])).squeeze() / \
        (forward[None, :] @ COVinv @ forward[:, None])


def get_target(obj):
    
    """Compute the target signal from EEG data using LCMV beamforming.

    This function computes the target signal from EEG data using Linearly Constrained Minimum Variance (LCMV) beamforming [1].
    It extracts the target signal by spatially filtering the EEG data based on the forward model.

    Parameters:
    -----------
    obj : RawCLAM or EpochsCLAM object
        The RawCLAM or EpochsCLAM object containing EEG data.

    Returns:
    --------
    target : ndarray
        The target signal extracted from the EEG data using LCMV beamforming.

    Raises:
    -------
    Exception:
        If the input object is not an instance of RawCLAM or EpochsCLAM.
        
    References:
    -----------
    [1] Van Veen, Barry D., and Kevin M. Buckley. "Beamforming: A versatile approach to spatial filtering." 
        IEEE assp magazine 5.2 (1988): 4-24.
    """
    
    if not (isinstance(obj, RawCLAM) or isinstance(obj, EpochsCLAM)):
        raise Exception('get_target can only be applied to RawCLAM or EpochsCLAM objects')
    ixs_goods = _get_ixs_goods(obj)
    target_codes = obj.marker_definition.keys()
    if isinstance(obj, mne.Epochs):
        epochs_events = obj.events
        epochs_data = obj.get_data(ixs_goods)
        forward_goods = obj.forward_full[ixs_goods]
        COV = np.mean([np.cov(np.real(np.concatenate(epochs_data[epochs_events[:, 2]
                      == target_code], axis=-1))) for target_code in target_codes], axis=0)
        w = _get_lcmv_weights(COV, forward_goods)
        target = np.array([w @ ep for ep in epochs_data]).squeeze()
    else:
        raw_events = mne.events_from_annotations(obj)[0]
        raw_data = obj.get_data(ixs_goods)
        epochs = EpochsCLAM(obj)
        epochs_events = epochs.events
        epochs_data = epochs.get_data(ixs_goods)
        forward_goods = epochs.forward_full[ixs_goods]
        COV = np.mean([np.cov(np.real(np.concatenate(epochs_data[epochs_events[:, 2]
                      == target_code], axis=-1))) for target_code in target_codes], axis=0)
        w = _get_lcmv_weights(COV, forward_goods)
        target = (w @ raw_data).squeeze()
    target *= obj.flip

    return target


def set_flip(obj, plot=False):
    
    """Determine and set the dipole sign flip for the target signal.

    This method determines the dipole sign flip for the target signal based on its waveform asymmetry.
    It analyzes the phases of the target signal to determine whether the rising or falling phase is dominant (longer),
    and sets the flip factor accordingly.

    Parameters:
    -----------
    obj : RawCLAM or EpochsCLAM object
        The RawCLAM or EpochsCLAM object containing EEG data.
    plot : bool, optional
        Whether to plot the distribution of target signal phases and the result of flip determination. Default is False.

    Returns:
    --------
    None

    Raises:
    -------
    Exception:
        If the input object is not an instance of RawCLAM or EpochsCLAM.
    """
    
    if not (isinstance(obj, RawCLAM) or isinstance(obj, EpochsCLAM)):
        raise Exception('set_flip can only be applied to RawCLAM or EpochsCLAM objects')
    obj.flip = 1 # required for get_target() to run
    target = get_target(obj)
    is_complex = np.iscomplexobj(target)
    if not is_complex:
        n_times = target.shape[-1]
        n_fft = next_fast_len(n_times)
        target = hilbert(target, N=n_fft, axis=-1)[..., :n_times]
    target_phases = np.angle(target)
    n_rising = (target_phases < 0).sum()
    n_falling = (target_phases > 0).sum()
    if n_rising > n_falling:
        flip = -1
    else:
        flip = 1
    if plot:
        p = binomtest(n_rising, n_rising + n_falling).pvalue
        _, ax = plt.subplots(1, 1, subplot_kw={'projection': 'polar'})
        ax.hist(target_phases.flatten(), color='k', alpha=0.3)
        ax.set_title('flip = {:d}, p = {:.4f}'.format(flip, p))
        ax.yaxis.grid(False)
        ax.xaxis.grid(False)
        ax.get_yaxis().set_visible(False)
        ax.axvline(circmean(target_phases.flatten()), c='r')
    obj.flip = flip


def set_forward(raw, l_freq_noise, h_freq_noise, n_comp=4):
    
    """Compute and set the forward model for target source reconstruction.

    This method computes and sets the forward model necessary for LCMV beamforming.
    A data-driven approach called Spatio-Spectral Decomposition (SSD) [1] is used to 
    find components in the signal that maximize power in the target frequency range
    while minimizing power in the noise frequency range.
    
    A component must be selected by the user by clicking on the respective power
    spectrum before closing the plot.

    Parameters:
    -----------
    raw : RawCLAM object
        The RawCLAM object containing EEG data.
    l_freq_noise : float
        Lower edge of the noise frequency range.
    h_freq_noise : float
        Higher edge of the noise frequency range.
    n_comp : int, optional
        Number of SSD components to plot. Default is 4.

    Returns:
    --------
    None

    Raises:
    -------
    Exception:
        If the input object is not an instance of RawCLAM.
        If the data does not contain frequencies of at least 1 - 40 Hz.
        If no target for stimulation (forward model) is selected by the user.
        
    References:
    -----------
    [1] Nikulin, Vadim V., Guido Nolte, and Gabriel Curio. "A novel method for reliable and fast extraction of neuronal 
        EEG/MEG oscillations on the basis of spatio-spectral decomposition." NeuroImage 55.4 (2011): 1528-1535.
    """
    
    if not isinstance(raw, RawCLAM):
        raise Exception('set_forward can only be applied to RawCLAM objects')
    sfreq = raw.info['sfreq']
    l_freq = raw.info['highpass']
    h_freq = raw.info['lowpass']
    l_freq_target = raw.l_freq_target
    h_freq_target = raw.h_freq_target
    marker_definition = raw.marker_definition
    participant = raw.participant
    ixs_goods = _get_ixs_goods(raw)
    if not (l_freq <= 1 and 40 <= h_freq):
        raise Exception(
            'Data should contain frequencies of at least 1 - 40 Hz')
    data_broad = raw.copy().filter(1, 40).get_data(ixs_goods)
    data_signal = raw.copy().filter(
        l_freq_target, h_freq_target).get_data(ixs_goods)
    data_noise = raw.copy().filter(
        l_freq_noise, h_freq_noise).get_data(ixs_goods)
    COV_B = np.cov(data_broad)
    COV_S = np.cov(data_signal)
    COV_N = np.cov(data_noise)
    evals, evecs = linalg.eig(COV_S, COV_N)
    ix = np.argsort(evals)[::-1]
    D = evecs[:, ix].T
    M = linalg.pinv(D)
    fig, axes = plt.subplots(
        n_comp, 2, figsize=(
            7, 7), gridspec_kw={
            'width_ratios': [
                1, 1]})
    for ix_comp in range(n_comp):
        w = _get_lcmv_weights(COV_B, M[:, ix_comp])
        psd, freqs = psd_array_welch(
            w @ data_broad, raw.info['sfreq'], fmin=1, fmax=40, n_fft=int(3 * raw.info['sfreq']))
        axes[ix_comp, 0].semilogy(freqs, psd.flatten(), c='k')
        axes[ix_comp, 0].tick_params(axis='x', labelsize=8)
        axes[ix_comp, 0].tick_params(axis='y', labelsize=5)
        axes[ix_comp, 0].axvline(
            l_freq_target, color='grey', linestyle='--', linewidth=0.5)
        axes[ix_comp, 0].axvline(
            h_freq_target, color='grey', linestyle='--', linewidth=0.5)
        
        plot_topomap(M[:, ix_comp], mne.pick_info(raw.info, ixs_goods), axes=axes[ix_comp,
                     1], sensors=False, contours=0, show=False)
    plt.suptitle(
        'Please select a target for stimulation by clicking anywhere inside the left plot')
    plt.figtext(0.25, 0, 'Frequency (Hz)')
    plt.figtext(0, 0.5, 'Power (a.u.)', rotation='vertical')
    plt.tight_layout()
    default_linewidth = axes[0, 0].spines['bottom'].get_linewidth()
    ix_comp = None

    def _onclick_ax(event, axes=axes[:, 0], fig=fig):
        nonlocal ix_comp
        ax_pressed = None
        ix_ax_pressed = None
        for ix, ax in enumerate(axes):
            if ax.contains(event)[0]:
                ax_pressed = ax
                ix_ax_pressed = ix
            else:
                ax.spines['top'].set_visible(False)
                ax.spines['right'].set_visible(False)
                ax.spines['bottom'].set_color('k')
                ax.spines['left'].set_color('k')
                ax.spines['bottom'].set_linewidth(default_linewidth)
                ax.spines['left'].set_linewidth(default_linewidth)
        if ax_pressed is not None:
            ix_comp = ix_ax_pressed
            ax_pressed.spines['top'].set_visible(True)
            ax_pressed.spines['right'].set_visible(True)
            ax_pressed.spines['top'].set_color('g')
            ax_pressed.spines['right'].set_color('g')
            ax_pressed.spines['bottom'].set_color('g')
            ax_pressed.spines['left'].set_color('g')
            ax_pressed.spines['top'].set_linewidth(2 * default_linewidth)
            ax_pressed.spines['right'].set_linewidth(2 * default_linewidth)
            ax_pressed.spines['bottom'].set_linewidth(2 * default_linewidth)
            ax_pressed.spines['left'].set_linewidth(2 * default_linewidth)
            fig.canvas.draw()
    fig.canvas.mpl_connect("button_press_event", _onclick_ax)
    sns.despine()
    plt.show()
    if ix_comp is None:
        raise Exception(
            'No target for stimulation (forward model) was selected')
    forward_full = np.zeros(raw.n_chs)
    forward_full[ixs_goods] = M[:, int(ix_comp)]
    raw.forward_full = forward_full

# from scipy.io import loadmat
# raw = mne.io.read_raw_brainvision('C:\\Users\\hasla\Desktop\\rising_falling_cwm_tims\\data\\P1_DH\\calibration_no_stim.vhdr',preload=True)
# raw = raw.pick_channels(raw.ch_names[:64])
# raw.set_montage('easycap-M1',match_case=False)
# forward_model = loadmat('C:\\Users\\hasla\Desktop\\rising_falling_cwm_tims\\data\\P1_DH\\P_TARGET_64.mat')['P_TARGET_64'].squeeze()
# flip = loadmat('C:\\Users\\hasla\Desktop\\rising_falling_cwm_tims\\data\\P1_DH\\flip.mat')['flip'].squeeze()
# sfreq = raw.info['sfreq']
# mask_bad = forward_model==0
# bads = np.array(raw.ch_names)[:64][mask_bad]
# raw.drop_channels(bads)
# get_forward(raw,8,14,1,40)


import mne
import numpy as np
from scipy import linalg
from mne.filter import next_fast_len
from scipy.signal import hilbert
from pycircstat.descriptive import mean as circmean
import matplotlib.pyplot as plt
import pandas as pd
from .stats import _wrap
import seaborn as sns
from scipy.stats import ttest_ind
import emd
from math import degrees
from scipy.stats import f_oneway
from pingouin import circ_corrcl
from fooof import FOOOF
from mne.time_frequency import psd_array_welch
from .misc import _get_ixs_goods, _get_main_target_phase
from .base import RawCLAM, EpochsCLAM
from .beamformer import get_target


# TODO:
# Brain - heart coupling


def compute_phase_tracking(raw, plot=False):
    
    """
    Compute the phase lag between CLAM-NIBS and targeted brain oscillations.

    Parameters:
    -----------
    raw : RawCLAM
        The RawCLAM object containing the raw data to analyze.
    plot : bool, optional
        Whether to plot the phase tracking results (default is False).

    Returns:
    --------
    float
        The mean differene between the targeted phase lag and the actual phase lag, 
        computed across all targeted phase lags.

    Raises:
    -------
    Exception
        If the input raw data is not of type RawCLAM.
        If the data is not filtered into the target frequency range.
        If the function is called on data without CLAM-NIBS.
    """
    
    if not isinstance(raw, RawCLAM):
        raise Exception('compute_phase_tracking can only be applied to RawCLAM objects')

    sfreq = raw.info['sfreq']
    l_freq = raw.info['highpass']
    h_freq = raw.info['lowpass']
    l_freq_target = raw.l_freq_target
    h_freq_target = raw.h_freq_target
    marker_definition = raw.marker_definition
    participant = raw.participant

    if not (l_freq == l_freq_target and h_freq == h_freq_target):
        raise Exception(
            'Data must be filtered into the target frequency range to compute phase tracking')
    if not raw.is_stim:
        raise Exception(
            'Phase tracking can only be computed for data with CLAM-tACS')

    df_phase = pd.DataFrame({'Target Phase': [], 'Actual Phase': []})
    target_codes = list(marker_definition.keys())
    target_phases = list(marker_definition.values())
    target_labels = ['{:d}'.format(int(degrees(x))) for x in target_phases]
    n_targets = len(target_codes)
    epochs_hil = EpochsCLAM(raw)
    target_hil = get_target(epochs_hil)
    envelope_hil = epochs_hil.get_data(['envelope']).squeeze()
    for target_code, target_label in zip(target_codes, target_labels):
        mask_condition = epochs_hil.events[:, 2] == target_code
        phasediffs = _wrap(
            np.angle(
                target_hil[mask_condition]) -
            np.angle(
                envelope_hil[mask_condition])).flatten()
        df_new = pd.DataFrame(
            {'Target Phase': [target_label] * len(phasediffs), 'Actual Phase': phasediffs})
        df_phase = pd.concat([df_phase, df_new])
    mean_actual_phases = [_wrap(circmean(df_phase.groupby('Target Phase').get_group(
        target_labels[ix])['Actual Phase'].to_numpy())) for ix in range(n_targets)]
    if plot:
        max_subplots_per_row = 3
        n_rows = n_targets // max_subplots_per_row + \
            min(1, n_targets % max_subplots_per_row)
        n_cols = min(n_targets, max_subplots_per_row)
        fig, axs = plt.subplots(n_rows, n_cols, figsize=(
            4 * n_rows, 4 * n_cols), subplot_kw={'projection': 'polar'})
        axs = axs.flatten()
        for ix in range(n_targets):
            axs[ix].axvline(
                mean_actual_phases[ix],
                c='b',
                label='Mean Actual Phase' if ix == 0 else None)
            axs[ix].axvline(
                target_phases[ix],
                c='g',
                label='Target Phase' if ix == 0 else None)
            axs[ix].hist(
                df_phase.groupby('Target Phase').get_group(
                    target_labels[ix])['Actual Phase'].to_numpy(),
                color='k',
                alpha=0.3)
            axs[ix].set_title(target_labels[ix])
            axs[ix].yaxis.grid(False)
            axs[ix].xaxis.grid(False)
            axs[ix].get_yaxis().set_visible(False)
        plt.legend()
        plt.suptitle('Actual Phase')
    phase_delay = circmean(
        [tp - ap for tp, ap in zip(target_phases, mean_actual_phases)])
    return phase_delay


def _get_fooof_amplitude(
        data,
        sfreq,
        l_freq,
        h_freq,
        l_freq_target,
        h_freq_target):
    psd, freqs = psd_array_welch(data, sfreq, l_freq, h_freq, int(2.5 * sfreq))
    fm = FOOOF(verbose=False)
    fm.fit(freqs, psd)
    peak_params = fm.get_results().peak_params
    peak_params = peak_params[(peak_params[:, 0] > l_freq_target) & (
        peak_params[:, 0] < h_freq_target)]
    amp = np.max(peak_params[:, 1])
    return amp


def _get_hht_amplitude(
        data,
        sfreq,
        l_freq,
        h_freq,
        l_freq_target,
        h_freq_target):
    imf = emd.sift.mask_sift(data, max_imfs=4)
    IP, IF, IA = emd.spectra.frequency_transform(imf, sfreq, 'hilbert')
    freqs_carrier, hht = emd.spectra.hilberthuang(
        IF, IA, (l_freq, h_freq, int(2.5 * (h_freq - l_freq))))
    amp = hht[(freqs_carrier > l_freq_target) & (
        freqs_carrier < h_freq_target)].max()
    return amp


def compute_amplitude_modulation(raw, measure='hilbert_amp'):
    
    """
    Compute the phase-lag dependent modulation of target oscillation amplitude by CLAM-NIBS.

    Parameters:
    -----------
    raw : RawCLAM
        The RawCLAM object containing the raw data to analyze.
    measure : str, optional
        The method used to compute amplitude modulation. It can be one of the following:
            - 'hht_amp': Amplitude estimation based on Hilbert-Huang Transform.
            - 'fooof_amp': Amplitude estimation based on FOOOF algorithm.
            - 'hilbert_amp': Amplitude estimation based on Hilbert transform (default).

    Returns:
    --------
    pandas.DataFrame
        A DataFrame containing the computed amplitude values for each epoch.

    Raises:
    -------
    Exception
        If the input raw data is not of type RawCLAM.
        If the method for computing amplitude is not supported.
        If the Raw object does not meet the requirements for the chosen method.
    """
    
    if not isinstance(raw, RawCLAM):
        raise Exception('compute_phase_tracking can only be applied to RawCLAM objects')

    sfreq = raw.info['sfreq']
    l_freq = raw.info['highpass']
    h_freq = raw.info['lowpass']
    l_freq_target = raw.l_freq_target
    h_freq_target = raw.h_freq_target
    marker_definition = raw.marker_definition
    participant = raw.participant
    session = raw.session
    design = raw.design
    events = mne.events_from_annotations(raw)[0]

    if measure not in ['hht_amp', 'fooof_amp', 'hilbert_amp']:
        raise Exception(
            'Method to compute amplitude must be \'hht_amp\', \'fooof_amp\', or \'hilbert_amp\'')
    if (l_freq > 1 or h_freq < 30) and measure in ['hht_amp', 'fooof_amp']:
        raise Exception(
            'Raw object must have a passband of at least 1 - 30 Hz for amplitude estimation based on HHT or FOOOF')
    if not (l_freq == l_freq_target and h_freq ==
            h_freq_target) and measure == 'hilbert_amp':
        raise Exception(
            'Raw object must be filtered into the target frequency range for amplitude estimation based on Hilbert')
    if design == 'trial_wise':
        epochs = EpochsCLAM(raw)
        target_hil = get_target(epochs)
    else:
        target_hil = get_target(raw.copy().apply_hilbert())
        target_hil = target_hil[None, :, :]
    epoch_amps = []
    for epoch_hil in target_hil:
        if measure == 'hht_amp':
            amp = _get_hht_amplitude(
                np.real(epoch_hil),
                sfreq,
                l_freq,
                h_freq,
                l_freq_target,
                h_freq_target)
        elif measure == 'fooof_amp':
            amp = _get_fooof_amplitude(
                np.real(epoch_hil),
                sfreq,
                l_freq,
                h_freq,
                l_freq_target,
                h_freq_target)
        else:
            amp = np.mean(np.abs(epoch_hil))
        epoch_amps.append(amp)
    if design == 'trial_wise':
        epoch_target_phases = np.vectorize(
            marker_definition.get)(epochs.events[:, 2])
    else:
        epoch_target_phases = [_get_main_target_phase(marker_definition, events)]
    df_result = pd.DataFrame({'participant': [participant] * len(epoch_amps),
                              'session': [session] * len(epoch_amps),
                              'design': [design] * len(epoch_amps),
                              'target_phase': epoch_target_phases,
                              'measure': [measure] * len(epoch_amps),
                              'value': epoch_amps})
    return df_result


def _get_fooof_exponent(
        data,
        sfreq,
        l_freq,
        h_freq,
        l_freq_target,
        h_freq_target):
    psd, freqs = psd_array_welch(data, sfreq, l_freq, h_freq, int(2.5 * sfreq))
    fm = FOOOF(verbose=False)
    fm.fit(freqs, psd)
    ap = fm.get_results().aperiodic_params[-1]
    return ap


def compute_aperiodic_modulation(raw, measure='fooof_ae'):
    
    """
    Compute the phase-lag dependent modulation of target aperiodic activity by CLAM-NIBS.

    Parameters:
    -----------
    raw : RawCLAM
        The RawCLAM object containing the raw data to analyze.
    measure : str, optional
        The method used to compute the aperiodic exponent. It can only be 'fooof_ae' (default).
    plot : bool, optional
        Whether to plot the aperiodic modulation results (default is False).

    Returns:
    --------
    pandas.DataFrame
        A DataFrame containing the computed aperiodic exponent for each epoch.

    Raises:
    -------
    Exception
        If the input raw data is not of type RawCLAM.
        If the method for computing the aperiodic exponent is not supported.
        If the Raw object does not meet the requirements for the chosen method.
    """
    
    if not isinstance(raw, RawCLAM):
        raise Exception('compute_aperiodic_modulation can only be applied to RawCLAM objects')

    sfreq = raw.info['sfreq']
    l_freq = raw.info['highpass']
    h_freq = raw.info['lowpass']
    l_freq_target = raw.l_freq_target
    h_freq_target = raw.h_freq_target
    marker_definition = raw.marker_definition
    participant = raw.participant
    session = raw.session
    design = raw.design
    events = mne.events_from_annotations(raw)[0]

    if measure not in ['fooof_ae']:
        raise Exception(
            'Method to compute aperiodic exponent must be \'fooof_ae\'')
    if (l_freq > 1 or h_freq < 30) and measure in ['fooof_ae']:
        raise Exception(
            'Raw object must have a passband of at least 1 - 30 Hz for aperiodic exponent estimation based on FOOOF')
    if design == 'trial_wise':
        epochs = EpochsCLAM(raw)
        target_hil = get_target(epochs)
    else:
        target_hil = get_target(raw.copy().apply_hilbert())
        target_hil = target_hil[None, :, :]
    epoch_aes = []
    for epoch_hil in target_hil:
        ap = _get_fooof_exponent(
            np.real(epoch_hil),
            sfreq,
            l_freq,
            h_freq,
            l_freq_target,
            h_freq_target)
        epoch_aes.append(ap)
    if design == 'trial_wise':
        epoch_target_phases = np.vectorize(
            marker_definition.get)(epochs.events[:, 2])
    else:
        epoch_target_phases = [_get_main_target_phase(marker_definition, events)]
    df_result = pd.DataFrame({'participant': [participant] * len(epoch_aes),
                              'session': [session] * len(epoch_aes),
                              'design': [design] * len(epoch_aes),
                              'target_phase': epoch_target_phases,
                              'measure': [measure] * len(epoch_aes),
                              'value': epoch_aes})
    return df_result


def _get_fooof_frequency(
        data,
        sfreq,
        l_freq,
        h_freq,
        l_freq_target,
        h_freq_target):
    psd, freqs = psd_array_welch(data, sfreq, l_freq, h_freq, int(2.5 * sfreq))
    fm = FOOOF(verbose=False)
    fm.fit(freqs, psd)
    peak_params = fm.get_results().peak_params
    peak_params = peak_params[(peak_params[:, 0] > l_freq_target) & (
        peak_params[:, 0] < h_freq_target)]
    pf = peak_params[np.argmax(peak_params[:, 1])][0]
    return pf


def compute_frequency_modulation(raw, measure='fooof_pf'):
    
    """
    Compute the phase-lag dependent modulation of target oscillation frequency by CLAM-NIBS.

    Parameters:
    -----------
    raw : RawCLAM
        The RawCLAM object containing the raw data to analyze.
    measure : str, optional
        The method used to compute the peak frequency. It can only be 'fooof_pf' (default).
    plot : bool, optional
        Whether to plot the frequency modulation results (default is False).

    Returns:
    --------
    pandas.DataFrame
        A DataFrame containing the computed peak frequency values for each epoch.

    Raises:
    -------
    Exception
        If the input raw data is not of type RawCLAM.
        If the method for computing the peak frequency is not supported.
        If the Raw object does not meet the requirements for the chosen method.
    """
    
    if not isinstance(raw, RawCLAM):
        raise Exception('compute_frequency_modulation can only be applied to RawCLAM objects')

    sfreq = raw.info['sfreq']
    l_freq = raw.info['highpass']
    h_freq = raw.info['lowpass']
    l_freq_target = raw.l_freq_target
    h_freq_target = raw.h_freq_target
    marker_definition = raw.marker_definition
    participant = raw.participant
    session = raw.session
    design = raw.design
    events = mne.events_from_annotations(raw)[0]

    if measure not in ['fooof_pf']:
        raise Exception(
            'Method to compute peak frequency must be \'fooof_pf\'')
    if (l_freq > 1 or h_freq < 30) and measure in ['fooof_pf']:
        raise Exception(
            'Raw object must have a passband of at least 1 - 30 Hz for peak frequency estimation based on FOOOF')
    if design == 'trial_wise':
        epochs = EpochsCLAM(raw)
        target_hil = get_target(epochs)
    else:
        target_hil = get_target(raw.copy().apply_hilbert())
        target_hil = target_hil[None, :, :]
    epoch_pfs = []
    for epoch_hil in target_hil:
        ap = _get_fooof_frequency(
            np.real(epoch_hil),
            sfreq,
            l_freq,
            h_freq,
            l_freq_target,
            h_freq_target)
        epoch_pfs.append(ap)
    if design == 'trial_wise':
        epoch_target_phases = np.vectorize(
            marker_definition.get)(epochs.events[:, 2])
    else:
        epoch_target_phases = [_get_main_target_phase(marker_definition, events)]
    df_result = pd.DataFrame({'participant': [participant] * len(epoch_pfs),
                              'session': [session] * len(epoch_pfs),
                              'design': [design] * len(epoch_pfs),
                              'target_phase': epoch_target_phases,
                              'measure': [measure] * len(epoch_pfs),
                              'value': epoch_pfs})
    return df_result